from __future__ import annotations

import asyncio
import contextlib
import json
import logging
from pprint import pprint
from typing import Any, AsyncIterator, Dict, TYPE_CHECKING
import sys

import aioredis
import click
from tabulate import tabulate

from ai.backend.common.cli import EnumChoice, MinMaxRange
from ai.backend.common.config import redis_config_iv
from ai.backend.common.docker import ImageRef
from ai.backend.common.etcd import (
    AsyncEtcd, ConfigScopes,
    quote as etcd_quote,
    unquote as etcd_unquote,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.gateway.config import SharedConfig
from ai.backend.gateway.defs import REDIS_IMAGE_DB
if TYPE_CHECKING:
    from .__main__ import CLIContext

log = BraceStyleAdapter(logging.getLogger(__name__))


@click.group()
def cli() -> None:
    pass


@contextlib.asynccontextmanager
async def etcd_ctx(cli_ctx: CLIContext) -> AsyncIterator[AsyncEtcd]:
    local_config = cli_ctx.local_config
    creds = None
    if local_config['etcd']['user']:
        creds = {
            'user': local_config['etcd']['user'],
            'password': local_config['etcd']['password'],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: '',
        # TODO: provide a way to specify other scope prefixes
    }
    etcd = AsyncEtcd(local_config['etcd']['addr'], local_config['etcd']['namespace'],
                     scope_prefix_map, credentials=creds)
    try:
        yield etcd
    finally:
        await etcd.close()


@contextlib.asynccontextmanager
async def config_ctx(cli_ctx: CLIContext) -> AsyncIterator[SharedConfig]:
    local_config = cli_ctx.local_config
    ctx: Dict[str, Any] = {}
    ctx['config'] = local_config
    # scope_prefix_map is created inside ConfigServer
    shared_config = SharedConfig(
        ctx,
        local_config['etcd']['addr'],
        local_config['etcd']['user'],
        local_config['etcd']['password'],
        local_config['etcd']['namespace'],
    )
    await shared_config.reload()
    raw_redis_config = await shared_config.etcd.get_prefix('config/redis')
    local_config['redis'] = redis_config_iv.check(raw_redis_config)
    redis_image = await aioredis.create_redis(
        local_config['redis']['addr'].as_sockaddr(),
        password=local_config['redis']['password'] if local_config['redis']['password'] else None,
        timeout=3.0,
        encoding='utf8',
        db=REDIS_IMAGE_DB)
    ctx['redis_image'] = redis_image
    try:
        yield shared_config
    finally:
        redis_image.close()
        await redis_image.wait_closed()
        await shared_config.close()


@cli.command()
@click.argument('key')
@click.argument('value')
@click.option('-s', '--scope', type=EnumChoice(ConfigScopes), default=ConfigScopes.GLOBAL,
              help='The configuration scope to put the value.')
@click.pass_obj
def put(cli_ctx: CLIContext, key, value, scope) -> None:
    '''Put a single key-value pair into the etcd.'''
    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                await etcd.put(key, value, scope=scope)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('key', type=str)
@click.argument('file', type=click.File('rb'))
@click.option('-s', '--scope', type=EnumChoice(ConfigScopes),
              default=ConfigScopes.GLOBAL,
              help='The configuration scope to put the value.')
@click.pass_obj
def put_json(cli_ctx: CLIContext, key, file, scope) -> None:
    '''
    Put a JSON object from FILE to the etcd as flattened key-value pairs
    under the given KEY prefix.
    '''
    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                value = json.load(file)
                await etcd.put_prefix(key, value, scope=scope)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('src_prefix', type=str)
@click.argument('dst_prefix', type=str)
@click.option('-s', '--scope', type=EnumChoice(ConfigScopes),
              default=ConfigScopes.GLOBAL,
              help='The configuration scope to get/put the subtree. '
                   'To move between different scopes, use the global scope '
                   'and specify the per-scope prefixes manually.')
@click.pass_obj
def move_subtree(cli_ctx: CLIContext, src_prefix, dst_prefix, scope) -> None:
    '''
    Move a subtree to another key prefix.
    '''
    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                subtree = await etcd.get_prefix(src_prefix, scope=scope)
                await etcd.put_prefix(dst_prefix, subtree, scope=scope)
                await etcd.delete_prefix(src_prefix, scope=scope)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('key')
@click.option('--prefix', is_flag=True,
              help='Get all key-value pairs prefixed with the given key '
                   'as a JSON form.')
@click.option('-s', '--scope', type=EnumChoice(ConfigScopes),
              default=ConfigScopes.GLOBAL,
              help='The configuration scope to put the value.')
@click.pass_obj
def get(cli_ctx: CLIContext, key, prefix, scope) -> None:
    '''
    Get the value of a key in the configured etcd namespace.
    '''
    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                if prefix:
                    data = await etcd.get_prefix(key, scope=scope)
                    print(json.dumps(dict(data), indent=4))
                else:
                    val = await etcd.get(key, scope=scope)
                    if val is None:
                        sys.exit(1)
                    print(val)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('key')
@click.option('--prefix', is_flag=True,
              help='Delete all keys prefixed with the given key.')
@click.option('-s', '--scope', type=EnumChoice(ConfigScopes),
              default=ConfigScopes.GLOBAL,
              help='The configuration scope to put the value.')
@click.pass_obj
def delete(cli_ctx: CLIContext, key, prefix, scope) -> None:
    '''Delete the key in the configured etcd namespace.'''
    async def _impl():
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                if prefix:
                    await etcd.delete_prefix(key, scope=scope)
                else:
                    await etcd.delete(key, scope=scope)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.option('-s', '--short', is_flag=True,
              help='Show only the image references and digests.')
@click.option('-i', '--installed', is_flag=True,
              help='Show only the installed images.')
@click.pass_obj
def list_images(cli_ctx: CLIContext, short, installed) -> None:
    '''List all configured images.'''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            displayed_items = []
            try:
                items = await shared_config.list_images()
                for item in items:
                    if installed and not item['installed']:
                        continue
                    if short:
                        img = ImageRef(f"{item['name']}:{item['tag']}",
                                       item['registry'])
                        displayed_items.append((img.canonical, item['digest']))
                    else:
                        pprint(item)
                if short:
                    print(tabulate(displayed_items, tablefmt='plain'))
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('reference')
@click.pass_obj
def inspect_image(cli_ctx: CLIContext, reference) -> None:
    '''Show the details of the given image or alias.'''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                item = await shared_config.inspect_image(reference)
                pprint(item)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('reference')
@click.pass_obj
def forget_image(cli_ctx: CLIContext, reference) -> None:
    '''
    Forget (delete) a specific image.
    NOTE: aliases to the given reference are NOT deleted.
    '''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                await shared_config.forget_image(reference)
                log.info('Done.')
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('reference')
@click.argument('slot_type')
@click.argument('range_value', type=MinMaxRange)
@click.pass_obj
def set_image_resource_limit(cli_ctx: CLIContext, reference, slot_type, range_value) -> None:
    '''Set the MIN:MAX values of a SLOT_TYPE limit for the given image REFERENCE.'''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                await shared_config.set_image_resource_limit(
                    reference, slot_type, range_value)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('registry')
@click.pass_obj
def rescan_images(cli_ctx: CLIContext, registry) -> None:
    '''
    Update the kernel image metadata from all configured docker registries.

    Pass the name (usually hostname or "lablup") of the Docker registry configured as REGISTRY.
    '''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                await shared_config.rescan_images(registry)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('alias')
@click.argument('target')
@click.pass_obj
def alias(cli_ctx: CLIContext, alias, target) -> None:
    '''Add an image alias from the given alias to the target image reference.'''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                await shared_config.alias(alias, target)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('alias')
@click.pass_obj
def dealias(cli_ctx: CLIContext, alias) -> None:
    '''Remove an alias.'''
    async def _impl():
        async with config_ctx(cli_ctx) as shared_config:
            try:
                await shared_config.dealias(alias)
            except Exception:
                log.exception('An error occurred.')
    with cli_ctx.logger:
        asyncio.run(_impl())


@cli.command()
@click.argument('value')
@click.pass_obj
def quote(cli_ctx: CLIContext, value) -> None:
    '''
    Quote the given string for use as a URL piece in etcd keys.
    Use this to generate argument inputs for aliases and raw image keys.
    '''
    print(etcd_quote(value))


@cli.command()
@click.argument('value')
@click.pass_obj
def unquote(cli_ctx: CLIContext, value) -> None:
    '''
    Unquote the given string used as a URL piece in etcd keys.
    '''
    print(etcd_unquote(value))



class InvalidArgument(Exception):
    """
    An internal exception class to represent invalid arguments in internal APIs.
    This is wrapped as InvalidAPIParameters in web request handlers.
    """
    pass



"""stats-refactor

Revision ID: 4cc87e7fbfdf
Revises: e18ed5fcfedf
Create Date: 2019-05-30 18:40:17.669756

"""
from datetime import timedelta
from decimal import Decimal
import math

from alembic import op
import sqlalchemy as sa
from ai.backend.manager.models.base import (
    convention, IDColumn, ResourceSlotColumn,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.expression import bindparam

# revision identifiers, used by Alembic.
revision = '4cc87e7fbfdf'
down_revision = 'e18ed5fcfedf'
branch_labels = None
depends_on = None


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)

    # previous table def used for migration
    kernels = sa.Table(
        'kernels', metadata,

        # preserved and referred columns
        IDColumn(),
        sa.Column('created_at', sa.DateTime(timezone=True),
                  server_default=sa.func.now(), index=True),
        sa.Column('terminated_at', sa.DateTime(timezone=True),
                  nullable=True, default=sa.null(), index=True),
        sa.Column('occupied_slots', ResourceSlotColumn(), nullable=False),
        sa.Column('occupied_shares', postgresql.JSONB(), nullable=False, default={}),

        # old column(s) to be migrated and removed
        sa.Column('cpu_used', sa.BigInteger(), default=0),       # msec
        sa.Column('mem_max_bytes', sa.BigInteger(), default=0),
        sa.Column('net_rx_bytes', sa.BigInteger(), default=0),
        sa.Column('net_tx_bytes', sa.BigInteger(), default=0),
        sa.Column('io_read_bytes', sa.BigInteger(), default=0),
        sa.Column('io_write_bytes', sa.BigInteger(), default=0),
        sa.Column('io_max_scratch_size', sa.BigInteger(), default=0),

        # new column(s) to be added
        sa.Column('last_stat', postgresql.JSONB(), nullable=True, default=sa.null()),
    )

    op.add_column('kernels', sa.Column('last_stat', postgresql.JSONB(astext_type=sa.Text()),
                                       nullable=True))

    connection = op.get_bind()
    query = sa.select([
        kernels.c.id,
        kernels.c.created_at,
        kernels.c.terminated_at,
        kernels.c.occupied_slots,
        kernels.c.occupied_shares,
        kernels.c.cpu_used,
        kernels.c.mem_max_bytes,
        kernels.c.net_rx_bytes,
        kernels.c.net_tx_bytes,
        kernels.c.io_read_bytes,
        kernels.c.io_write_bytes,
        kernels.c.io_max_scratch_size,
    ]).select_from(kernels).order_by(kernels.c.created_at)
    results = connection.execute(query).fetchall()

    updates = []
    q_pct = Decimal('0.00')
    for row in results:
        if row['terminated_at'] is None:
            cpu_avg_pct = 0
        else:
            cpu_avg_pct = (
                Decimal(100) *
                Decimal(timedelta(microseconds=1e3 * row['cpu_used']) /
                        (row['terminated_at'] - row['created_at']))
            ).quantize(q_pct).normalize()
        mem_capacity = 0
        _oslots = row['occupied_slots']
        if _oslots:
            mem_capacity = _oslots.get('mem')
        if mem_capacity is None or mem_capacity == 0:
            # fallback: try legacy field
            _oshares = row['occupied_shares']
            mem_capacity = _oshares.get('mem')
        if mem_capacity is None or mem_capacity == 0:
            # fallback: round-up to nearest GiB
            mem_capacity = math.ceil(row['mem_max_bytes'] / (2**30)) * (2**30)
        if mem_capacity is None or mem_capacity == 0:
            # fallback: assume 1 GiB
            mem_capacity = 2**30
        last_stat = {
            'cpu_used': {
                'current': str(row['cpu_used']),
                'capacity': None,
            },
            'cpu_util': {
                'current': str(cpu_avg_pct),
                'capacity': None,
                'stats.avg': str(cpu_avg_pct),
            },
            'mem': {
                'current': str(row['mem_max_bytes']),
                'capacity': str(mem_capacity),
                'stats.max': str(row['mem_max_bytes']),
            },
            'io_read': {
                'current': str(row['io_read_bytes']),
                'capacity': None,
                'stats.rate': '0',
            },
            'io_write': {
                'current': str(row['io_write_bytes']),
                'capacity': None,
                'stats.rate': '0',
            },
            'io_scratch_size': {
                'current': str(row['io_max_scratch_size']),
                'capacity': str(10 * (2**30)),  # 10 GiB
                'stats.max': str(row['io_max_scratch_size']),
            },
        }
        updates.append({'row_id': row['id'], 'last_stat': last_stat})

    if updates:
        query = (sa.update(kernels)
                   .values(last_stat=bindparam('last_stat'))
                   .where(kernels.c.id == bindparam('row_id')))
        connection.execute(query, updates)

    op.drop_column('kernels', 'io_max_scratch_size')
    op.drop_column('kernels', 'net_rx_bytes')
    op.drop_column('kernels', 'net_tx_bytes')
    op.drop_column('kernels', 'mem_max_bytes')
    op.drop_column('kernels', 'io_write_bytes')
    op.drop_column('kernels', 'io_read_bytes')
    op.drop_column('kernels', 'cpu_used')


def downgrade():
    metadata = sa.MetaData(naming_convention=convention)

    kernels = sa.Table(
        'kernels', metadata,

        # preserved and referred columns
        IDColumn(),
        sa.Column('created_at', sa.DateTime(timezone=True),
                  server_default=sa.func.now(), index=True),
        sa.Column('terminated_at', sa.DateTime(timezone=True),
                  nullable=True, default=sa.null(), index=True),
        sa.Column('occupied_slots', ResourceSlotColumn(), nullable=False),
        sa.Column('occupied_shares', postgresql.JSONB(), nullable=False, default={}),

        # old column(s) to be migrated
        sa.Column('last_stat', postgresql.JSONB(), nullable=True, default=sa.null()),

        # new column(s) to be added
        sa.Column('cpu_used', sa.BigInteger(), default=0),       # msec
        sa.Column('mem_max_bytes', sa.BigInteger(), default=0),
        sa.Column('net_rx_bytes', sa.BigInteger(), default=0),
        sa.Column('net_tx_bytes', sa.BigInteger(), default=0),
        sa.Column('io_read_bytes', sa.BigInteger(), default=0),
        sa.Column('io_write_bytes', sa.BigInteger(), default=0),
        sa.Column('io_max_scratch_size', sa.BigInteger(), default=0),
    )

    op.add_column('kernels', sa.Column('cpu_used', sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('io_read_bytes', sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('io_write_bytes', sa.BIGINT(),
                                       autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('mem_max_bytes', sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('net_tx_bytes', sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('net_rx_bytes', sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column('kernels', sa.Column('io_max_scratch_size', sa.BIGINT(),
                                       autoincrement=False, nullable=True))

    # Restore old stats
    connection = op.get_bind()
    query = sa.select([kernels.c.id, kernels.c.last_stat]).select_from(kernels)
    results = connection.execute(query).fetchall()
    updates = []
    for row in results:
        last_stat = row['last_stat']
        updates.append({
            'row_id': row['id'],
            'cpu_used': Decimal(last_stat['cpu_used']['current']),
            'io_read_bytes': int(last_stat['io_read']['current']),
            'io_write_bytes': int(last_stat['io_write']['current']),
            'mem_max_bytes': int(last_stat['mem']['stats.max']),
            'io_max_scratch_size': int(last_stat['io_scratch_size']['stats.max']),
        })
    if updates:
        query = (sa.update(kernels)
                 .values({
                     'cpu_used': bindparam('cpu_used'),
                     'io_read_bytes': bindparam('io_read_bytes'),
                     'io_write_bytes': bindparam('io_write_bytes'),
                     'mem_max_bytes': bindparam('mem_max_bytes'),
                     'net_tx_bytes': 0,
                     'net_rx_bytes': 0,
                     'io_max_scratch_size': bindparam('io_max_scratch_size'),
                 })
                 .where(kernels.c.id == bindparam('row_id')))
        connection.execute(query, updates)

    op.drop_column('kernels', 'last_stat')



"""add_bootstrap_script_to_keypairs

Revision ID: ed666f476f39
Revises: d643752544de
Create Date: 2020-03-15 17:40:46.754121

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed666f476f39'
down_revision = 'd643752544de'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('keypairs', sa.Column('bootstrap_script',
                                        sa.String(length=64 * 1024),
                                        nullable=False,
                                        server_default=''))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('keypairs', 'bootstrap_script')
    # ### end Alembic commands ###



import asyncio
from contextlib import asynccontextmanager as actxmgr
import itertools
from typing import (
    Any,
    AsyncIterator,
    Final,
    Iterable,
    Mapping,
    Tuple,
)
from uuid import UUID

import aiohttp
import attr
import yarl

from ..exceptions import InvalidArgument


@attr.s(auto_attribs=True, slots=True, frozen=True)
class StorageProxyInfo:
    session: aiohttp.ClientSession
    secret: str
    client_api_url: yarl.URL
    manager_api_url: yarl.URL


AUTH_TOKEN_HDR: Final = 'X-BackendAI-Storage-Auth-Token'


class StorageSessionManager:

    _proxies: Mapping[str, StorageProxyInfo]

    def __init__(self, storage_config: Mapping[str, Any]) -> None:
        self.config = storage_config
        self._proxies = {}
        for proxy_name, proxy_config in self.config['proxies'].items():
            connector = aiohttp.TCPConnector(ssl=proxy_config['ssl_verify'])
            session = aiohttp.ClientSession(connector=connector)
            self._proxies[proxy_name] = StorageProxyInfo(
                session=session,
                secret=proxy_config['secret'],
                client_api_url=yarl.URL(proxy_config['client_api']),
                manager_api_url=yarl.URL(proxy_config['manager_api']),
            )

    async def aclose(self) -> None:
        close_aws = []
        for proxy_info in self._proxies.values():
            close_aws.append(proxy_info.session.close())
        await asyncio.gather(*close_aws, return_exceptions=True)

    @staticmethod
    def split_host(vfolder_host: str) -> Tuple[str, str]:
        proxy_name, _, volume_name = vfolder_host.partition(':')
        return proxy_name, volume_name

    async def get_all_volumes(self) -> Iterable[Tuple[str, Mapping[str, Any]]]:
        fetch_aws = []

        async def _fetch(
            proxy_name: str,
            proxy_info: StorageProxyInfo,
        ) -> Iterable[Tuple[str, Mapping[str, Any]]]:
            async with proxy_info.session.request(
                'GET', proxy_info.manager_api_url / 'volumes',
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return ((proxy_name, volume_data) for volume_data in reply['volumes'])

        for proxy_name, proxy_info in self._proxies.items():
            fetch_aws.append(_fetch(proxy_name, proxy_info))
        results = await asyncio.gather(*fetch_aws)
        return itertools.chain(*results)

    async def get_mount_path(self, vfolder_host: str, vfolder_id: UUID) -> str:
        async with self.request(
            vfolder_host, 'GET', 'folder/mount',
            json={
                'volume': self.split_host(vfolder_host)[1],
                'vfid': str(vfolder_id),
            },
        ) as (_, resp):
            reply = await resp.json()
            return reply['path']

    @actxmgr
    async def request(
        self,
        vfolder_host_or_proxy_name: str,
        method: str,
        request_relpath: str,
        /,
        *args,
        **kwargs,
    ) -> AsyncIterator[Tuple[yarl.URL, aiohttp.ClientResponse]]:
        proxy_name, _ = self.split_host(vfolder_host_or_proxy_name)
        try:
            proxy_info = self._proxies[proxy_name]
        except KeyError:
            raise InvalidArgument('There is no such storage proxy', proxy_name)
        headers = kwargs.pop('headers', {})
        headers[AUTH_TOKEN_HDR] = proxy_info.secret
        async with proxy_info.session.request(
            method, proxy_info.manager_api_url / request_relpath,
            *args,
            headers=headers,
            **kwargs,
        ) as client_resp:
            yield proxy_info.client_api_url, client_resp



from abc import ABCMeta, abstractmethod
from typing import (
    Tuple,
    Sequence,
)

from aiohttp import web

from ai.backend.common.plugin import AbstractPlugin, BasePluginContext
from ...gateway.types import CORSOptions, WebMiddleware


class WebappPlugin(AbstractPlugin, metaclass=ABCMeta):
    """
    Webapp plugins should create a valid aiohttp.web.Application instance.  The returned app
    instance will be a subapp of the root app defined by the manager, and additional user-properties
    will be set as defined in ``ai.backend.gateway.server.PUBLIC_INTERFACES``.

    The init/cleanup methods of the plugin are ignored and the manager uses the standard aiohttp's
    application lifecycle handlers attached to the returned app instance.
    """

    @abstractmethod
    async def create_app(
        self, cors_options: CORSOptions,
    ) -> Tuple[web.Application, Sequence[WebMiddleware]]:
        pass


class WebappPluginContext(BasePluginContext[WebappPlugin]):
    plugin_group = 'backendai_webapp_v20'



import enum
from typing import (
    Any,
    List,
    Mapping,
    Sequence,
)

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
import trafaret as t

from ai.backend.common import validators as tx
from ai.backend.common.types import SessionTypes

from ..defs import DEFAULT_ROLE
from ..exceptions import InvalidArgument
from .base import metadata, GUID, IDColumn, EnumType
from .user import UserRole
from .vfolder import verify_vfolder_name

__all__: Sequence[str] = (
    'TemplateType', 'session_templates', 'query_accessible_session_templates'
)


class TemplateType(str, enum.Enum):
    TASK = 'task'
    CLUSTER = 'cluster'


session_templates = sa.Table(
    'session_templates', metadata,
    IDColumn('id'),
    sa.Column('created_at', sa.DateTime(timezone=True), index=True),
    sa.Column('is_active', sa.Boolean, default=True),

    sa.Column('domain_name', sa.String(length=64), sa.ForeignKey('domains.name'), nullable=False),
    sa.Column('group_id', GUID, sa.ForeignKey('groups.id'), nullable=True),
    sa.Column('user_uuid', GUID, sa.ForeignKey('users.uuid'), index=True, nullable=False),
    sa.Column('type', EnumType(TemplateType), nullable=False, server_default='TASK', index=True),

    sa.Column('name', sa.String(length=128), nullable=True),
    sa.Column('template', pgsql.JSONB(), nullable=False)
)


task_template_v1 = t.Dict({
    tx.AliasedKey(['api_version', 'apiVersion']): t.String,
    t.Key('kind'): t.Enum('taskTemplate', 'task_template'),
    t.Key('metadata'): t.Dict({
        t.Key('name'): t.String,
        t.Key('tag', default=None): t.Null | t.String,
    }),
    t.Key('spec'): t.Dict({
        tx.AliasedKey(['type', 'sessionType'],
                      default='interactive') >> 'session_type': tx.Enum(SessionTypes),
        t.Key('kernel'): t.Dict({
            t.Key('image'): t.String,
            t.Key('environ', default={}): t.Null | t.Mapping(t.String, t.String),
            t.Key('run', default=None): t.Null | t.Dict({
                t.Key('bootstrap', default=None): t.Null | t.String,
                tx.AliasedKey(['startup', 'startup_command', 'startupCommand'],
                              default=None) >> 'startup_command': t.Null | t.String
            }),
            t.Key('git', default=None): t.Null | t.Dict({
                t.Key('repository'): t.String,
                t.Key('commit', default=None): t.Null | t.String,
                t.Key('branch', default=None): t.Null | t.String,
                t.Key('credential', default=None): t.Null | t.Dict({
                    t.Key('username'): t.String,
                    t.Key('password'): t.String
                }),
                tx.AliasedKey(['destination_dir', 'destinationDir'],
                              default=None) >> 'dest_dir': t.Null | t.String
            })
        }),
        t.Key('mounts', default={}): t.Null | t.Mapping(t.String, t.Any),
        t.Key('resources', default=None): t.Null | t.Mapping(t.String, t.Any)
    })
}).allow_extra('*')


def check_task_template(raw_data: Mapping[str, Any]) -> Mapping[str, Any]:
    data = task_template_v1.check(raw_data)
    if mounts := data['spec'].get('mounts'):
        for p in mounts.values():
            if p is None:
                continue
            if not p.startswith('/home/work/'):
                raise InvalidArgument(f'Path {p} should start with /home/work/')
            if not verify_vfolder_name(p.replace('/home/work/', '')):
                raise InvalidArgument(f'Path {p} is reserved for internal operations.')
    return data


cluster_template_v1 = t.Dict({
    tx.AliasedKey(['api_version', 'apiVersion']): t.String,
    t.Key('kind'): t.Enum('clusterTemplate', 'cluster_template'),
    t.Key('mode'): t.Enum('single-node', 'multi-node'),
    t.Key('metadata'): t.Dict({
        t.Key('name'): t.String
    }),
    t.Key('spec'): t.Dict({
        t.Key('environ', default={}): t.Null | t.Mapping(t.String, t.String),
        t.Key('mounts', default={}): t.Null | t.Mapping(t.String, t.Any),
        t.Key('nodes'): t.List(t.Dict({
            t.Key('role'): t.String,
            tx.AliasedKey(['session_template', 'sessionTemplate']): tx.UUID,
            t.Key('replicas', default=1): t.Int
        }))
    })
}).allow_extra('*')


def check_cluster_template(raw_data: Mapping[str, Any]) -> Mapping[str, Any]:
    data = cluster_template_v1.check(raw_data)
    defined_roles: List[str] = []
    for node in data['spec']['nodes']:
        node['session_template'] = str(node['session_template'])
        if node['role'] in defined_roles:
            raise InvalidArgument("Each role can only be defined once")
        if node['role'] == DEFAULT_ROLE and node['replicas'] != 1:
            raise InvalidArgument(
                f"One and only one {DEFAULT_ROLE} node must be created per cluster"
            )
        defined_roles.append(node['role'])
    if DEFAULT_ROLE not in defined_roles:
        raise InvalidArgument(
            f"One and only one {DEFAULT_ROLE} node must be created per cluster"
        )
    return data


async def query_accessible_session_templates(conn, user_uuid, template_type: TemplateType, *,
                                             user_role=None, domain_name=None,
                                             allowed_types=['user'],
                                             extra_conds=None):
    from ai.backend.manager.models import groups, users, association_groups_users as agus
    entries = []
    if 'user' in allowed_types:
        # Query user templates
        j = (session_templates.join(users, session_templates.c.user_uuid == users.c.uuid))
        query = (sa.select([
                        session_templates.c.name,
                        session_templates.c.id,
                        session_templates.c.created_at,
                        session_templates.c.user_uuid,
                        session_templates.c.group_id,
                        users.c.email
                    ])
                    .select_from(j)
                    .where((session_templates.c.user_uuid == user_uuid) &
                           session_templates.c.is_active &
                           (session_templates.c.type == template_type)))
        if extra_conds is not None:
            query = query.where(extra_conds)
        result = await conn.execute(query)
        async for row in result:
            entries.append({
                'name': row.name,
                'id': row.id,
                'created_at': row.created_at,
                'is_owner': True,
                'user': str(row.user_uuid) if row.user_uuid else None,
                'group': str(row.group_id) if row.group_id else None,
                'user_email': row.email,
                'group_name': None,
            })
    if 'group' in allowed_types:
        # Query group session_templates
        if user_role == UserRole.ADMIN or user_role == 'admin':
            query = (sa.select([groups.c.id])
                        .select_from(groups)
                        .where(groups.c.domain_name == domain_name))
            result = await conn.execute(query)
            grps = await result.fetchall()
            group_ids = [g.id for g in grps]
        else:
            j = sa.join(agus, users, agus.c.user_id == users.c.uuid)
            query = (sa.select([agus.c.group_id])
                        .select_from(j)
                        .where(agus.c.user_id == user_uuid))
            result = await conn.execute(query)
            grps = await result.fetchall()
            group_ids = [g.group_id for g in grps]
        j = (session_templates.join(groups, session_templates.c.group_id == groups.c.id))
        query = (sa.select([
                        session_templates.c.name,
                        session_templates.c.id,
                        session_templates.c.created_at,
                        session_templates.c.user_uuid,
                        session_templates.c.group_id,
                        groups.c.name
                    ], use_labels=True)
                    .where(session_templates.c.group_id.in_(group_ids) &
                           session_templates.c.is_active &
                           (session_templates.c.type == template_type)))
        if extra_conds is not None:
            query = query.where(extra_conds)
        if 'user' in allowed_types:
            query = query.where(session_templates.c.user_uuid != user_uuid)
        result = await conn.execute(query)
        is_owner = (user_role == UserRole.ADMIN or user_role == 'admin')
        async for row in result:
            entries.append({
                'name': row.session_templates_name,
                'id': row.session_templates_id,
                'created_at': row.session_templates_created_at,
                'is_owner': is_owner,
                'user': (str(row.session_templates_user_uuid) if row.session_templates_user_uuid
                         else None),
                'group': str(row.session_templates_group_id) if row.session_templates_group_id else None,
                'user_email': None,
                'group_name': row.groups_name,
            })
    return entries



import asyncio
import enum
import functools
import json
import logging
from typing import FrozenSet
import sqlalchemy as sa
import trafaret as t
from typing import (
    Any, Final,
    Iterable,
    Tuple,
)

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
from aiotools import aclosing

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common import validators as tx

from . import ManagerStatus
from .auth import superadmin_required
from .exceptions import (
    InstanceNotFound,
    InvalidAPIParameters,
    GenericBadRequest,
    ServerFrozen,
    ServiceUnavailable,
)
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params
from ..manager.defs import DEFAULT_ROLE
from ..manager.models import agents, kernels, AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.manager'))


class SchedulerOps(enum.Enum):
    INCLUDE_AGENTS = 'include-agents'
    EXCLUDE_AGENTS = 'exclude-agents'


def server_status_required(allowed_status: FrozenSet[ManagerStatus]):

    def decorator(handler):

        @functools.wraps(handler)
        async def wrapped(request, *args, **kwargs):
            status = await request.app['shared_config'].get_manager_status()
            if status not in allowed_status:
                if status == ManagerStatus.FROZEN:
                    raise ServerFrozen
                msg = f'Server is not in the required status: {allowed_status}'
                raise ServiceUnavailable(msg)
            return (await handler(request, *args, **kwargs))

        return wrapped

    return decorator


READ_ALLOWED: Final = frozenset({ManagerStatus.RUNNING, ManagerStatus.FROZEN})
ALL_ALLOWED: Final = frozenset({ManagerStatus.RUNNING})


class GQLMutationUnfrozenRequiredMiddleware:

    def resolve(self, next, root, info, **args):
        if info.operation.operation == 'mutation' and \
                info.context['manager_status'] == ManagerStatus.FROZEN:
            raise ServerFrozen
        return next(root, info, **args)


async def detect_status_update(app):
    try:
        async with aclosing(app['shared_config'].watch_manager_status()) as agen:
            async for ev in agen:
                if ev.event == 'put':
                    app['shared_config'].get_manager_status.cache_clear()
                    updated_status = await app['shared_config'].get_manager_status()
                    log.debug('Process-{0} detected manager status update: {1}',
                              app['pidx'], updated_status)
    except asyncio.CancelledError:
        pass


@atomic
async def fetch_manager_status(request: web.Request) -> web.Response:
    log.info('MANAGER.FETCH_MANAGER_STATUS ()')
    try:
        status = await request.app['shared_config'].get_manager_status()
        etcd_info = await request.app['shared_config'].get_manager_nodes_info()
        configs = request.app['local_config']['manager']

        async with request.app['dbpool'].acquire() as conn, conn.begin():
            query = (sa.select([sa.func.count(kernels.c.id)])
                       .select_from(kernels)
                       .where((kernels.c.cluster_role == DEFAULT_ROLE) &
                              (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES))))
            active_sessions_num = await conn.scalar(query)

            # TODO: update logic to return information for multiple managers (HA)
            if '' in etcd_info:
                _id = etcd_info['']
            elif etcd_info:
                _id = list(etcd_info.keys())[0]
            else:
                _id = ''
            nodes = [
                {
                    'id': _id,
                    'num_proc': configs['num-proc'],
                    'service_addr': str(configs['service-addr']),
                    'heartbeat_timeout': configs['heartbeat-timeout'],
                    'ssl_enabled': configs['ssl-enabled'],
                    'active_sessions': active_sessions_num,
                    'status': status.value,
                }
            ]
            return web.json_response({
                'nodes': nodes,
                'status': status.value,                  # legacy?
                'active_sessions': active_sessions_num,  # legacy?
            })
    except:
        log.exception('GET_MANAGER_STATUS: exception')
        raise


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('status'): tx.Enum(ManagerStatus, use_name=True),
        t.Key('force_kill', default=False): t.ToBool,
    }))
async def update_manager_status(request: web.Request, params: Any) -> web.Response:
    log.info('MANAGER.UPDATE_MANAGER_STATUS (status:{}, force_kill:{})',
             params['status'], params['force_kill'])
    try:
        params = await request.json()
        status = params['status']
        force_kill = params['force_kill']
    except json.JSONDecodeError:
        raise InvalidAPIParameters(extra_msg='No request body!')
    except (AssertionError, ValueError) as e:
        raise InvalidAPIParameters(extra_msg=str(e.args[0]))

    if force_kill:
        await request.app['registry'].kill_all_sessions()
    await request.app['shared_config'].update_manager_status(status)

    return web.Response(status=204)


@atomic
async def get_announcement(request: web.Request) -> web.Response:
    data = await request.app['shared_config'].etcd.get('manager/announcement')
    if data is None:
        ret = {'enabled': False, 'message': ''}
    else:
        ret = {'enabled': True, 'message': data}
    return web.json_response(ret)


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('enabled', default='false'): t.ToBool,
        t.Key('message', default=None): t.Null | t.String,
    }))
async def update_announcement(request: web.Request, params: Any) -> web.Response:
    if params['enabled']:
        if not params['message']:
            raise InvalidAPIParameters(extra_msg='Empty message not allowed to enable announcement')
        await request.app['shared_config'].etcd.put('manager/announcement', params['message'])
    else:
        await request.app['shared_config'].etcd.delete('manager/announcement')
    return web.Response(status=204)


iv_scheduler_ops_args = {
    SchedulerOps.INCLUDE_AGENTS: t.List(t.String),
    SchedulerOps.EXCLUDE_AGENTS: t.List(t.String),
}


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('op'): tx.Enum(SchedulerOps),
        t.Key('args'): t.Any,
    }))
async def perform_scheduler_ops(request: web.Request, params: Any) -> web.Response:
    try:
        args = iv_scheduler_ops_args[params['op']].check(params['args'])
    except t.DataError as e:
        raise InvalidAPIParameters(
            f"Input validation failed for args with {params['op']}",
            extra_data=e.as_dict(),
        )
    if params['op'] in (SchedulerOps.INCLUDE_AGENTS, SchedulerOps.EXCLUDE_AGENTS):
        schedulable = (params['op'] == SchedulerOps.INCLUDE_AGENTS)
        async with request.app['dbpool'].acquire() as conn, conn.begin():
            query = (
                agents.update()
                .values(schedulable=schedulable)
                .where(agents.c.id.in_(args))
            )
            result = await conn.execute(query)
            if result.rowcount < len(args):
                raise InstanceNotFound()
        if schedulable:
            # trigger scheduler
            await request.app['event_dispatcher'].produce_event('do_schedule')
    else:
        raise GenericBadRequest('Unknown scheduler operation')
    return web.Response(status=204)


async def init(app: web.Application) -> None:
    app['status_watch_task'] = asyncio.create_task(detect_status_update(app))


async def shutdown(app: web.Application) -> None:
    if app['status_watch_task'] is not None:
        app['status_watch_task'].cancel()
        await app['status_watch_task']


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app['api_versions'] = (2, 3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    status_resource = cors.add(app.router.add_resource('/status'))
    cors.add(status_resource.add_route('GET', fetch_manager_status))
    cors.add(status_resource.add_route('PUT', update_manager_status))
    announcement_resource = cors.add(app.router.add_resource('/announcement'))
    cors.add(announcement_resource.add_route('GET', get_announcement))
    cors.add(announcement_resource.add_route('POST', update_announcement))
    cors.add(app.router.add_route('POST', '/scheduler/operation', perform_scheduler_ops))
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    return app, []



"""add inviter field for vfolder_invitations

Revision ID: 1fa6a31ea8e3
Revises: 26d0c387e764
Create Date: 2018-07-05 00:09:35.230704

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1fa6a31ea8e3'
down_revision = '26d0c387e764'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('vfolder_invitations', sa.Column('inviter', sa.String(length=256)))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('vfolder_invitations', 'inviter')
    # ### end Alembic commands ###



from __future__ import annotations

from collections import OrderedDict
from datetime import datetime
from decimal import Decimal
import enum
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypedDict,
    TypeVar,
    Union,
)
from uuid import UUID

from aiopg.sa.connection import SAConnection
from aiopg.sa.result import RowProxy
from aioredis import Redis
import graphene
from graphene.types.datetime import DateTime as GQLDateTime
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql

from ai.backend.common import msgpack, redis
from ai.backend.common.types import (
    AccessKey,
    BinarySize,
    ClusterMode,
    KernelId,
    SessionId,
    SessionTypes,
    SessionResult,
    SlotName,
    ResourceSlot,
)

from ..defs import DEFAULT_ROLE
from .base import (
    BigInt,
    EnumType,
    GUID,
    Item,
    KernelIDColumn,
    PaginatedList,
    ResourceSlotColumn,
    SessionIDColumnType,
    batch_result,
    batch_multiresult,
    metadata,
)
from .group import groups
from .user import users
from .keypair import keypairs

__all__: Sequence[str] = (
    'kernels',
    'session_dependencies',
    'KernelStatus',
    'ComputeContainer',
    'ComputeSession',
    'ComputeContainerList',
    'ComputeSessionList',
    'LegacyComputeSession',
    'LegacyComputeSessionList',
    'AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES',
    'USER_RESOURCE_OCCUPYING_KERNEL_STATUSES',
    'RESOURCE_USAGE_KERNEL_STATUSES',
    'DEAD_KERNEL_STATUSES',
    'LIVE_STATUS',
    'recalc_concurrency_used',
)


class KernelStatus(enum.Enum):
    # values are only meaningful inside the gateway
    PENDING = 0
    # ---
    PREPARING = 10
    # ---
    BUILDING = 20
    PULLING = 21
    # ---
    RUNNING = 30
    RESTARTING = 31
    RESIZING = 32
    SUSPENDED = 33
    # ---
    TERMINATING = 40
    TERMINATED = 41
    ERROR = 42
    CANCELLED = 43


# statuses to consider when calculating current resource usage
AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES = tuple(
    e for e in KernelStatus
    if e not in (
        KernelStatus.TERMINATED,
        KernelStatus.PENDING,
        KernelStatus.CANCELLED,
    )
)

USER_RESOURCE_OCCUPYING_KERNEL_STATUSES = tuple(
    e for e in KernelStatus
    if e not in (
        KernelStatus.TERMINATING,
        KernelStatus.TERMINATED,
        KernelStatus.PENDING,
        KernelStatus.CANCELLED,
    )
)

# statuses to consider when calculating historical resource usage
RESOURCE_USAGE_KERNEL_STATUSES = (
    KernelStatus.TERMINATED,
    KernelStatus.RUNNING,
)

DEAD_KERNEL_STATUSES = (
    KernelStatus.CANCELLED,
    KernelStatus.TERMINATED,
)

LIVE_STATUS = (
    KernelStatus.RUNNING,
)


def default_hostname(context) -> str:
    params = context.get_current_parameters()
    return f"{params['cluster_role']}{params['cluster_idx']}"


kernels = sa.Table(
    'kernels', metadata,
    # The Backend.AI-side UUID for each kernel
    # (mapped to a container in the docker backend and a pod in the k8s backend)
    KernelIDColumn(),
    # session_id == id when the kernel is the main container in a multi-container session or a
    # single-container session.
    # Otherwise, it refers the kernel ID of the main contaienr of the belonged multi-container session.
    sa.Column('session_id', SessionIDColumnType, unique=False, index=True, nullable=False),
    sa.Column('session_name', sa.String(length=64), unique=False, index=True),     # previously sess_id
    sa.Column('session_type', EnumType(SessionTypes), index=True, nullable=False,  # previously sess_type
              default=SessionTypes.INTERACTIVE, server_default=SessionTypes.INTERACTIVE.name),
    sa.Column('cluster_mode', sa.String(length=16), nullable=False,
              default=ClusterMode.SINGLE_NODE, server_default=ClusterMode.SINGLE_NODE.name),
    sa.Column('cluster_size', sa.Integer, nullable=False, default=1),
    sa.Column('cluster_role', sa.String(length=16), nullable=False, default=DEFAULT_ROLE),
    sa.Column('cluster_idx', sa.Integer, nullable=False, default=0),
    sa.Column('cluster_hostname', sa.String(length=64), nullable=False, default=default_hostname),

    # Resource ownership
    sa.Column('scaling_group', sa.ForeignKey('scaling_groups.name'), index=True, nullable=True),
    sa.Column('agent', sa.String(length=64), sa.ForeignKey('agents.id'), nullable=True),
    sa.Column('agent_addr', sa.String(length=128), nullable=True),
    sa.Column('domain_name', sa.String(length=64), sa.ForeignKey('domains.name'), nullable=False),
    sa.Column('group_id', GUID, sa.ForeignKey('groups.id'), nullable=False),
    sa.Column('user_uuid', GUID, sa.ForeignKey('users.uuid'), nullable=False),
    sa.Column('access_key', sa.String(length=20), sa.ForeignKey('keypairs.access_key')),
    sa.Column('image', sa.String(length=512)),
    sa.Column('registry', sa.String(length=512)),
    sa.Column('tag', sa.String(length=64), nullable=True),

    # Resource occupation
    sa.Column('container_id', sa.String(length=64)),
    sa.Column('occupied_slots', ResourceSlotColumn(), nullable=False),
    sa.Column('occupied_shares', pgsql.JSONB(), nullable=False, default={}),  # legacy
    sa.Column('environ', sa.ARRAY(sa.String), nullable=True),
    sa.Column('mounts', sa.ARRAY(sa.String), nullable=True),  # list of list
    sa.Column('mount_map', pgsql.JSONB(), nullable=True, default={}),
    sa.Column('attached_devices', pgsql.JSONB(), nullable=True, default={}),
    sa.Column('resource_opts', pgsql.JSONB(), nullable=True, default={}),
    sa.Column('bootstrap_script', sa.String(length=16 * 1024), nullable=True),

    # Port mappings
    # If kernel_host is NULL, it is assumed to be same to the agent host or IP.
    sa.Column('kernel_host', sa.String(length=128), nullable=True),
    sa.Column('repl_in_port', sa.Integer(), nullable=False),
    sa.Column('repl_out_port', sa.Integer(), nullable=False),
    sa.Column('stdin_port', sa.Integer(), nullable=False),   # legacy for stream_pty
    sa.Column('stdout_port', sa.Integer(), nullable=False),  # legacy for stream_pty
    sa.Column('service_ports', pgsql.JSONB(), nullable=True),
    sa.Column('preopen_ports', sa.ARRAY(sa.Integer), nullable=True),

    # Lifecycle
    sa.Column('created_at', sa.DateTime(timezone=True),
              server_default=sa.func.now(), index=True),
    sa.Column('terminated_at', sa.DateTime(timezone=True),
              nullable=True, default=sa.null(), index=True),
    sa.Column('starts_at', sa.DateTime(timezone=True),
              nullable=True, default=sa.null()),
    sa.Column('status', EnumType(KernelStatus),
              default=KernelStatus.PENDING,
              server_default=KernelStatus.PENDING.name,
              nullable=False, index=True),
    sa.Column('status_changed', sa.DateTime(timezone=True), nullable=True, index=True),
    sa.Column('status_info', sa.Unicode(), nullable=True, default=sa.null()),
    sa.Column('startup_command', sa.Text, nullable=True),
    sa.Column('result', EnumType(SessionResult),
              default=SessionResult.UNDEFINED,
              server_default=SessionResult.UNDEFINED.name,
              nullable=False, index=True),
    sa.Column('internal_data', pgsql.JSONB(), nullable=True),
    sa.Column('container_log', sa.LargeBinary(), nullable=True),
    # Resource metrics measured upon termination
    sa.Column('num_queries', sa.BigInteger(), default=0),
    sa.Column('last_stat', pgsql.JSONB(), nullable=True, default=sa.null()),

    sa.Index('ix_kernels_sess_id_role', 'session_id', 'cluster_role', unique=False),
    sa.Index('ix_kernels_updated_order',
             sa.func.greatest('created_at', 'terminated_at', 'status_changed'),
             unique=False),
    sa.Index('ix_kernels_unique_sess_token', 'access_key', 'session_id',
             unique=True,
             postgresql_where=sa.text(
                 "status NOT IN ('TERMINATED', 'CANCELLED') and "
                 "cluster_role = 'main'")),
)

session_dependencies = sa.Table(
    'session_dependencies', metadata,
    sa.Column('session_id', GUID,
              sa.ForeignKey('kernels.id', onupdate='CASCADE', ondelete='CASCADE'),
              index=True, nullable=False),
    sa.Column('depends_on', GUID,
              sa.ForeignKey('kernels.id', onupdate='CASCADE', ondelete='CASCADE'),
              index=True, nullable=False),
    sa.PrimaryKeyConstraint('session_id', 'depends_on'),
)

DEFAULT_SESSION_ORDERING = [
    sa.desc(sa.func.greatest(
        kernels.c.created_at,
        kernels.c.terminated_at,
        kernels.c.status_changed,
    ))
]


class SessionInfo(TypedDict):
    session_id: SessionId
    session_name: str
    status: KernelStatus
    created_at: datetime


async def match_session_ids(
    session_name_or_id: Union[str, UUID],
    access_key: AccessKey,
    *,
    db_connection: SAConnection,
    extra_cond=None,
    for_update: bool = False,
    max_matches: int = 10,
) -> Sequence[SessionInfo]:
    """
    Match the prefix of session ID or session name among the sessions that belongs to the given
    access key, and return the list of session IDs with matching prefixes.
    """
    cond_id = (
        (sa.sql.expression.cast(kernels.c.id, sa.String).like(f'{session_name_or_id}%')) &
        (kernels.c.access_key == access_key)
    )
    if extra_cond is not None:
        cond_id = cond_id & extra_cond
    cond_name = (
        (kernels.c.session_name.like(f'{session_name_or_id}%')) &
        (kernels.c.access_key == access_key)
    )
    if extra_cond is not None:
        cond_name = cond_name & extra_cond
    cond_session_id = (
        (sa.sql.expression.cast(kernels.c.session_id, sa.String).like(f'{session_name_or_id}%')) &
        (kernels.c.access_key == access_key)
    )
    if extra_cond is not None:
        cond_session_id = cond_session_id & extra_cond
    info_cols = [
        kernels.c.session_id,
        kernels.c.session_name,
        kernels.c.status,
        kernels.c.created_at,
    ]
    match_sid_by_id = (
        sa.select(info_cols, for_update=for_update)
        .select_from(kernels)
        .where(
            (kernels.c.session_id.in_(
                sa.select(
                    [kernels.c.session_id]
                )
                .select_from(kernels)
                .where(cond_id)
                .group_by(kernels.c.session_id)
                .limit(max_matches).offset(0)
            )) &
            (kernels.c.cluster_role == DEFAULT_ROLE)
        )
        .order_by(sa.desc(kernels.c.created_at))
    )
    match_sid_by_name = (
        sa.select(info_cols, for_update=for_update)
        .select_from(kernels)
        .where(
            (kernels.c.session_id.in_(
                sa.select(
                    [kernels.c.session_id]
                )
                .select_from(kernels)
                .where(cond_name)
                .group_by(kernels.c.session_id)
                .limit(max_matches).offset(0)
            )) &
            (kernels.c.cluster_role == DEFAULT_ROLE)
        )
        .order_by(sa.desc(kernels.c.created_at))
    )
    match_sid_by_session_id = (
        sa.select(info_cols, for_update=for_update)
        .select_from(kernels)
        .where(
            (kernels.c.session_id.in_(
                sa.select(
                    [kernels.c.session_id]
                )
                .select_from(kernels)
                .where(cond_session_id)
                .group_by(kernels.c.session_id)
                .limit(max_matches).offset(0)
            )) &
            (kernels.c.cluster_role == DEFAULT_ROLE)
        )
        .order_by(sa.desc(kernels.c.created_at))
    )
    for match_query in [
        match_sid_by_id,
        match_sid_by_session_id,
        match_sid_by_name,
    ]:
        result = await db_connection.execute(match_query)
        if result.rowcount == 0:
            continue
        return [
            SessionInfo(
                session_id=row['session_id'],
                session_name=row['session_name'],
                status=row['status'],
                created_at=row['created_at'],
            ) for row in await result.fetchall()
        ]
    return []


async def get_main_kernels(
    session_ids: Sequence[SessionId],
    *,
    db_connection: SAConnection,
    for_update: bool = False,
) -> Sequence[RowProxy]:
    """
    Return a list of the main kernels for the given session IDs.
    If a given session ID does not exist, its position will be ``None``.
    """
    session_id_to_rows = OrderedDict(
        (session_id, None) for session_id in session_ids
    )
    query = (
        sa.select([kernels])
        .select_from(kernels)
        .where(
            (kernels.c.session_id.in_(session_ids)) &
            (kernels.c.cluster_role == DEFAULT_ROLE)
        )
    )
    result = await db_connection.execute(query)
    for row in await result.fetchall():
        session_id_to_rows[row['session_id']] = row
    return [*session_id_to_rows.values()]


async def get_all_kernels(
    session_ids: Sequence[SessionId],
    *,
    db_connection: SAConnection,
    for_update: bool = False,
) -> Sequence[Sequence[RowProxy]]:
    """
    Return a list of all belonging kernel lists per the given session IDs
    in the order they are given.
    If a given session ID does not exist, an empty list will be returned
    at the position of that session ID.
    """
    session_id_to_rowsets: Dict[SessionId, List[RowProxy]]
    session_id_to_rowsets = OrderedDict(
        (session_id, []) for session_id in session_ids
    )
    for session_id in session_ids:
        query = (
            sa.select([sa.text('*')])
            .select_from(kernels)
            .where(
                (kernels.c.session_id == session_id)
            )
        )
        result = await db_connection.execute(query)
        if result.rowcount == 0:
            continue
        session_id_to_rowsets[session_id].extend(
            row for row in await result.fetchall()
        )
    return [*session_id_to_rowsets.values()]


class ComputeContainer(graphene.ObjectType):
    class Meta:
        interfaces = (Item, )

    # identity
    idx = graphene.Int()          # legacy
    role = graphene.String()      # legacy
    hostname = graphene.String()  # legacy
    cluster_idx = graphene.Int()
    cluster_role = graphene.String()
    cluster_hostname = graphene.String()
    session_id = graphene.UUID()  # owner session

    # image
    image = graphene.String()
    registry = graphene.String()

    # status
    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()

    # resources
    agent = graphene.String()
    container_id = graphene.String()
    resource_opts = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    live_stat = graphene.JSONString()
    last_stat = graphene.JSONString()

    @classmethod
    def parse_row(cls, context: Any, row: RowProxy) -> Mapping[str, Any]:
        assert row is not None
        from .user import UserRole
        is_superadmin = (context['user']['role'] == UserRole.SUPERADMIN)
        if is_superadmin:
            hide_agents = False
        else:
            hide_agents = context['local_config']['manager']['hide-agents']
        return {
            # identity
            'id': row['id'],
            'idx': row['cluster_idx'],
            'role': row['cluster_role'],
            'hostname': row['cluster_hostname'],
            'cluster_idx': row['cluster_idx'],
            'cluster_role': row['cluster_role'],
            'cluster_hostname': row['cluster_hostname'],
            'session_id': row['session_id'],

            # image
            'image': row['image'],
            'registry': row['registry'],

            # status
            'status': row['status'].name,
            'status_changed': row['status_changed'],
            'status_info': row['status_info'],
            'created_at': row['created_at'],
            'terminated_at': row['terminated_at'],
            'starts_at': row['starts_at'],
            'occupied_slots': row['occupied_slots'].to_json(),

            # resources
            'agent': row['agent'] if not hide_agents else None,
            'container_id': row['container_id'] if not hide_agents else None,
            'resource_opts': row['resource_opts'],

            # statistics
            # live_stat is resolved by Graphene
            'last_stat': row['last_stat'],
        }

    @classmethod
    def from_row(cls, context: Any, row: RowProxy) -> Optional[ComputeContainer]:
        if row is None:
            return None
        props = cls.parse_row(context, row)
        return cls(**props)

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        if not hasattr(self, 'status'):
            return None
        rs = info.context['redis_stat']
        if KernelStatus[self.status] in LIVE_STATUS:
            raw_live_stat = await redis.execute_with_retries(
                lambda: rs.get(str(self.id), encoding=None))
            if raw_live_stat is not None:
                live_stat = msgpack.unpackb(raw_live_stat)
                return live_stat
            return None
        else:
            return self.last_stat

    @classmethod
    async def load_count(
        cls,
        context,
        session_id: SessionId,
        *,
        cluster_role=None,
        domain_name=None,
        group_id=None,
        access_key=None,
    ) -> int:
        async with context['dbpool'].acquire() as conn:
            query = (
                sa.select([sa.func.count(kernels.c.id)])
                .select_from(kernels)
                .where(kernels.c.session_id == session_id)
                .as_scalar()
            )
            if cluster_role is not None:
                query = query.where(kernels.c.cluster_role == cluster_role)
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            result = await conn.execute(query)
            return await result.scalar()

    @classmethod
    async def load_slice(
        cls,
        context,
        limit: int,
        offset: int,
        session_id: SessionId,
        *,
        cluster_role=None,
        domain_name=None,
        group_id=None,
        access_key=None,
        order_key=None,
        order_asc=None,
    ) -> Sequence[Optional[ComputeContainer]]:
        async with context['dbpool'].acquire() as conn:
            if order_key is None:
                _ordering = DEFAULT_SESSION_ORDERING
            else:
                _order_func = sa.asc if order_asc else sa.desc
                _ordering = [_order_func(getattr(kernels.c, order_key))]
            j = (
                kernels
                .join(groups, groups.c.id == kernels.c.group_id)
                .join(users, users.c.uuid == kernels.c.user_uuid)
            )
            query = (
                sa.select([kernels])
                .select_from(j)
                .where(kernels.c.session_id == session_id)
                .order_by(*_ordering)
                .limit(limit)
                .offset(offset)
            )
            if cluster_role is not None:
                query = query.where(kernels.c.cluster_role == cluster_role)
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            return [cls.from_row(context, r) async for r in conn.execute(query)]

    @classmethod
    async def batch_load_by_session(
        cls,
        context,
        session_ids: Sequence[SessionId],
    ) -> Sequence[Sequence[ComputeContainer]]:
        async with context['dbpool'].acquire() as conn:
            query = (
                sa.select([kernels])
                .select_from(kernels)
                # TODO: use "owner session ID" when we implement multi-container session
                .where(kernels.c.session_id.in_(session_ids))
            )
            return await batch_multiresult(
                context, conn, query, cls,
                session_ids, lambda row: row['session_id'],
            )

    @classmethod
    async def batch_load_detail(
        cls,
        context,
        container_ids: Sequence[KernelId],
        *,
        domain_name=None,
        access_key=None,
    ) -> Sequence[Optional[ComputeContainer]]:
        async with context['dbpool'].acquire() as conn:
            j = (
                kernels
                .join(groups, groups.c.id == kernels.c.group_id)
                .join(users, users.c.uuid == kernels.c.user_uuid)
            )
            query = (
                sa.select([kernels])
                .select_from(j)
                .where(
                    (kernels.c.id.in_(container_ids))
                ))
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            return await batch_result(
                context, conn, query, cls,
                container_ids, lambda row: row['id'],
            )


class ComputeSession(graphene.ObjectType):
    class Meta:
        interfaces = (Item, )

    # identity
    tag = graphene.String()
    name = graphene.String()
    type = graphene.String()
    session_id = graphene.UUID()

    # image
    image = graphene.String()     # image for the main container
    registry = graphene.String()  # image registry for the main container
    cluster_template = graphene.String()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()

    # ownership
    domain_name = graphene.String()
    group_name = graphene.String()
    group_id = graphene.UUID()
    user_email = graphene.String()
    user_id = graphene.UUID()
    access_key = graphene.String()
    created_user_email = graphene.String()
    created_user_id = graphene.UUID()

    # status
    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()
    startup_command = graphene.String()
    result = graphene.String()

    # resources
    resource_opts = graphene.JSONString()
    scaling_group = graphene.String()
    service_ports = graphene.JSONString()
    mounts = graphene.List(lambda: graphene.String)
    occupied_slots = graphene.JSONString()

    # statistics
    num_queries = BigInt()

    # owned containers (aka kernels)
    containers = graphene.List(lambda: ComputeContainer)

    # relations
    dependencies = graphene.List(lambda: ComputeSession)

    @classmethod
    def parse_row(cls, context: Any, row: RowProxy) -> Mapping[str, Any]:
        assert row is not None
        return {
            # identity
            'id': row['id'],
            'tag': row['tag'],
            'name': row['session_name'],
            'type': row['session_type'].name,
            'session_id': row['session_id'],

            # image
            'image': row['image'],
            'registry': row['registry'],
            'cluster_template': None,  # TODO: implement
            'cluster_mode': row['cluster_mode'],
            'cluster_size': row['cluster_size'],

            # ownership
            'domain_name': row['domain_name'],
            'group_name': row['group_name'],
            'group_id': row['group_id'],
            'user_email': row['email'],
            'user_id': row['user_uuid'],
            'access_key': row['access_key'],
            'created_user_email': None,  # TODO: implement
            'created_user_id': None,     # TODO: implement

            # status
            'status': row['status'].name,
            'status_changed': row['status_changed'],
            'status_info': row['status_info'],
            'created_at': row['created_at'],
            'terminated_at': row['terminated_at'],
            'starts_at': row['starts_at'],
            'startup_command': row['startup_command'],
            'result': row['result'].name,

            # resources
            'resource_opts': row['resource_opts'],
            'scaling_group': row['scaling_group'],
            'service_ports': row['service_ports'],
            'mounts': row['mounts'],

            # statistics
            'num_queries': row['num_queries'],
        }

    @classmethod
    def from_row(cls, context: Mapping[str, Any], row: RowProxy) -> Optional[ComputeSession]:
        if row is None:
            return None
        props = cls.parse_row(context, row)
        return cls(**props)

    async def resolve_occupied_slots(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        """
        Calculate the sum of occupied resource slots of all sub-kernels,
        and return the JSON-serializable object from the sum result.
        """
        manager = info.context['dlmgr']
        loader = manager.get_loader('ComputeContainer.by_session')
        containers = await loader.load(self.session_id)
        zero = ResourceSlot()
        return sum(
            (ResourceSlot({
                SlotName(k): Decimal(v) for k, v in c.occupied_slots.items()
            }) for c in containers),
            start=zero,
        ).to_json()

    async def resolve_containers(
        self,
        info: graphene.ResolveInfo,
    ) -> Iterable[ComputeContainer]:
        manager = info.context['dlmgr']
        loader = manager.get_loader('ComputeContainer.by_session')
        return await loader.load(self.session_id)

    async def resolve_dependencies(
        self,
        info: graphene.ResolveInfo,
    ) -> Iterable[ComputeSession]:
        manager = info.context['dlmgr']
        loader = manager.get_loader('ComputeSession.by_dependency')
        return await loader.load(self.id)

    @classmethod
    async def load_count(
        cls,
        context,
        *,
        domain_name=None,
        group_id=None,
        access_key=None,
        status=None,
    ) -> int:
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(',')]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        async with context['dbpool'].acquire() as conn:
            query = (
                sa.select([sa.func.count(kernels.c.id)])
                .select_from(kernels)
                .where(kernels.c.cluster_role == DEFAULT_ROLE)
                .as_scalar()
            )
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            if status is not None:
                query = query.where(kernels.c.status.in_(status_list))
            result = await conn.execute(query)
            return await result.scalar()

    @classmethod
    async def load_slice(
        cls,
        context,
        limit: int,
        offset: int,
        *,
        domain_name=None,
        group_id=None,
        access_key=None,
        status=None,
        order_key=None,
        order_asc=None,
    ) -> Sequence[Optional[ComputeSession]]:
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(',')]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        async with context['dbpool'].acquire() as conn:
            if order_key is None:
                _ordering = DEFAULT_SESSION_ORDERING
            else:
                _order_func = sa.asc if order_asc else sa.desc
                _ordering = [_order_func(getattr(kernels.c, order_key))]
            j = (
                kernels
                .join(groups, groups.c.id == kernels.c.group_id)
                .join(users, users.c.uuid == kernels.c.user_uuid)
            )
            query = (
                sa.select([
                    kernels,
                    groups.c.name.label('group_name'),
                    users.c.email,
                ])
                .select_from(j)
                .where(kernels.c.cluster_role == DEFAULT_ROLE)
                .order_by(*_ordering)
                .limit(limit)
                .offset(offset)
            )
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            if status is not None:
                query = query.where(kernels.c.status.in_(status_list))
            return [cls.from_row(context, r) async for r in conn.execute(query)]

    @classmethod
    async def batch_load_by_dependency(
        cls,
        context,
        session_ids: Sequence[SessionId],
    ) -> Sequence[Sequence[ComputeSession]]:
        async with context['dbpool'].acquire() as conn:
            j = sa.join(
                kernels, session_dependencies,
                kernels.c.session_id == session_dependencies.c.depends_on,
            )
            query = (
                sa.select([kernels])
                .select_from(j)
                .where(
                    (kernels.c.cluster_role == DEFAULT_ROLE) &
                    (session_dependencies.c.session_id.in_(session_ids))
                )
            )
            return await batch_multiresult(
                context, conn, query, cls,
                session_ids, lambda row: row['id'],
            )

    @classmethod
    async def batch_load_detail(
        cls,
        context,
        session_ids: Sequence[SessionId],
        *,
        domain_name=None,
        access_key=None,
    ) -> Sequence[Optional[ComputeSession]]:
        async with context['dbpool'].acquire() as conn:
            j = (
                kernels
                .join(groups, groups.c.id == kernels.c.group_id)
                .join(users, users.c.uuid == kernels.c.user_uuid)
            )
            query = (
                sa.select([
                    kernels,
                    groups.c.name.label('group_name'),
                    users.c.email,
                ])
                .select_from(j)
                .where(
                    (kernels.c.cluster_role == DEFAULT_ROLE) &
                    (kernels.c.id.in_(session_ids))
                ))
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            return await batch_result(
                context, conn, query, cls,
                session_ids, lambda row: row['id'],
            )


class ComputeContainerList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(ComputeContainer, required=True)


class ComputeSessionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(ComputeSession, required=True)


# --------- pre-v5 legacy -----------

MetricValueType = TypeVar('MetricValueType', int, float)


class LegacyComputeSession(graphene.ObjectType):
    """
    Represents a main session.
    """
    class Meta:
        interfaces = (Item, )

    tag = graphene.String()  # Only for ComputeSession
    sess_id = graphene.String()    # legacy
    sess_type = graphene.String()  # legacy
    session_name = graphene.String()
    session_type = graphene.String()
    role = graphene.String()
    image = graphene.String()
    registry = graphene.String()
    domain_name = graphene.String()
    group_name = graphene.String()
    group_id = graphene.UUID()
    scaling_group = graphene.String()
    user_uuid = graphene.UUID()
    access_key = graphene.String()

    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    startup_command = graphene.String()
    result = graphene.String()

    # hidable fields by configuration
    agent = graphene.String()
    container_id = graphene.String()

    service_ports = graphene.JSONString()

    occupied_slots = graphene.JSONString()
    occupied_shares = graphene.JSONString()
    mounts = graphene.List(lambda: graphene.List(lambda: graphene.String))
    resource_opts = graphene.JSONString()

    num_queries = BigInt()
    live_stat = graphene.JSONString()
    last_stat = graphene.JSONString()

    user_email = graphene.String()

    # Legacy fields
    lang = graphene.String()
    mem_slot = graphene.Int()
    cpu_slot = graphene.Float()
    gpu_slot = graphene.Float()
    tpu_slot = graphene.Float()
    cpu_used = BigInt()
    cpu_using = graphene.Float()
    mem_max_bytes = BigInt()
    mem_cur_bytes = BigInt()
    net_rx_bytes = BigInt()
    net_tx_bytes = BigInt()
    io_read_bytes = BigInt()
    io_write_bytes = BigInt()
    io_max_scratch_size = BigInt()
    io_cur_scratch_size = BigInt()

    @classmethod
    async def _resolve_live_stat(
        cls,
        redis_stat: Redis,
        kernel_id: str,
    ) -> Optional[Mapping[str, Any]]:
        cstat = await redis.execute_with_retries(
            lambda: redis_stat.get(kernel_id, encoding=None))
        if cstat is not None:
            cstat = msgpack.unpackb(cstat)
        return cstat

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        if not hasattr(self, 'status'):
            return None
        rs = info.context['redis_stat']
        if KernelStatus[self.status] not in LIVE_STATUS:
            return self.last_stat
        else:
            return await type(self)._resolve_live_stat(rs, str(self.id))

    async def _resolve_legacy_metric(
        self,
        info: graphene.ResolveInfo,
        metric_key: str,
        metric_field: str,
        convert_type: Type[MetricValueType],
    ) -> Optional[MetricValueType]:
        if not hasattr(self, 'status'):
            return None
        rs = info.context['redis_stat']
        if KernelStatus[self.status] not in LIVE_STATUS:
            if self.last_stat is None:
                return convert_type(0)
            metric = self.last_stat.get(metric_key)
            if metric is None:
                return convert_type(0)
            value = metric.get(metric_field)
            if value is None:
                return convert_type(0)
            return convert_type(value)
        else:
            kstat = await type(self)._resolve_live_stat(rs, str(self.id))
            if kstat is None:
                return convert_type(0)
            metric = kstat.get(metric_key)
            if metric is None:
                return convert_type(0)
            value = metric.get(metric_field)
            if value is None:
                return convert_type(0)
            return convert_type(value)

    async def resolve_cpu_used(self, info: graphene.ResolveInfo) -> Optional[float]:
        return await self._resolve_legacy_metric(info, 'cpu_used', 'current', float)

    async def resolve_cpu_using(self, info: graphene.ResolveInfo) -> Optional[float]:
        return await self._resolve_legacy_metric(info, 'cpu_util', 'pct', float)

    async def resolve_mem_max_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'mem', 'stats.max', int)

    async def resolve_mem_cur_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'mem', 'current', int)

    async def resolve_net_rx_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'net_rx', 'stats.rate', int)

    async def resolve_net_tx_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'net_tx', 'stats.rate', int)

    async def resolve_io_read_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'io_read', 'current', int)

    async def resolve_io_write_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'io_write', 'current', int)

    async def resolve_io_max_scratch_size(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'io_scratch_size', 'stats.max', int)

    async def resolve_io_cur_scratch_size(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, 'io_scratch_size', 'current', int)

    @classmethod
    def parse_row(cls, context: Any, row: RowProxy) -> Mapping[str, Any]:
        assert row is not None
        from .user import UserRole
        mega = 2 ** 20
        is_superadmin = (context['user']['role'] == UserRole.SUPERADMIN)
        if is_superadmin:
            hide_agents = False
        else:
            hide_agents = context['local_config']['manager']['hide-agents']
        return {
            'id': row['id'],
            'sess_id': row['session_name'],         # legacy, will be deprecated
            'sess_type': row['session_type'].name,  # legacy, will be deprecated
            'session_name': row['session_name'],
            'session_type': row['session_type'].name,
            'role': row['role'],
            'tag': row['tag'],
            'image': row['image'],
            'registry': row['registry'],
            'domain_name': row['domain_name'],
            'group_name': row['name'],  # group.name (group is omitted since use_labels=True is not used)
            'group_id': row['group_id'],
            'scaling_group': row['scaling_group'],
            'user_uuid': row['user_uuid'],
            'access_key': row['access_key'],
            'status': row['status'].name,
            'status_changed': row['status_changed'],
            'status_info': row['status_info'],
            'created_at': row['created_at'],
            'terminated_at': row['terminated_at'],
            'startup_command': row['startup_command'],
            'result': row['result'].name,
            'service_ports': row['service_ports'],
            'occupied_slots': row['occupied_slots'].to_json(),
            'mounts': row['mounts'],
            'resource_opts': row['resource_opts'],
            'num_queries': row['num_queries'],
            # optionally hidden
            'agent': row['agent'] if not hide_agents else None,
            'container_id': row['container_id'] if not hide_agents else None,
            # live_stat is resolved by Graphene
            'last_stat': row['last_stat'],
            'user_email': row['email'],
            # Legacy fields
            # NOTE: currently graphene always uses resolve methods!
            'cpu_used': 0,
            'mem_max_bytes': 0,
            'mem_cur_bytes': 0,
            'net_rx_bytes': 0,
            'net_tx_bytes': 0,
            'io_read_bytes': 0,
            'io_write_bytes': 0,
            'io_max_scratch_size': 0,
            'io_cur_scratch_size': 0,
            'lang': row['image'],
            'occupied_shares': row['occupied_shares'],
            'mem_slot': BinarySize.from_str(
                row['occupied_slots'].get('mem', 0)) // mega,
            'cpu_slot': float(row['occupied_slots'].get('cpu', 0)),
            'gpu_slot': float(row['occupied_slots'].get('cuda.device', 0)),
            'tpu_slot': float(row['occupied_slots'].get('tpu.device', 0)),
        }

    @classmethod
    def from_row(cls, context: Any, row: RowProxy) -> Optional[LegacyComputeSession]:
        if row is None:
            return None
        props = cls.parse_row(context, row)
        return cls(**props)

    @classmethod
    async def load_count(cls, context, *,
                         domain_name=None, group_id=None, access_key=None,
                         status=None):
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(',')]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        async with context['dbpool'].acquire() as conn:
            query = (
                sa.select([sa.func.count(kernels.c.session_id)])
                .select_from(kernels)
                .where(kernels.c.cluster_role == DEFAULT_ROLE)
                .as_scalar()
            )
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            if status is not None:
                query = query.where(kernels.c.status.in_(status_list))
            result = await conn.execute(query)
            return await result.scalar()

    @classmethod
    async def load_slice(cls, context, limit, offset, *,
                         domain_name=None, group_id=None, access_key=None,
                         status=None,
                         order_key=None, order_asc=None):
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(',')]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        async with context['dbpool'].acquire() as conn:
            if order_key is None:
                _ordering = DEFAULT_SESSION_ORDERING
            else:
                _order_func = sa.asc if order_asc else sa.desc
                _ordering = [_order_func(getattr(kernels.c, order_key))]
            j = (kernels.join(groups, groups.c.id == kernels.c.group_id)
                        .join(users, users.c.uuid == kernels.c.user_uuid))
            query = (
                sa.select([kernels, groups.c.name, users.c.email])
                .select_from(j)
                .where(kernels.c.cluster_role == DEFAULT_ROLE)
                .order_by(*_ordering)
                .limit(limit)
                .offset(offset)
            )
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            if status is not None:
                query = query.where(kernels.c.status.in_(status_list))
            return [cls.from_row(context, r) async for r in conn.execute(query)]

    @classmethod
    async def batch_load(cls, context, access_keys, *,
                         domain_name=None, group_id=None,
                         status=None):
        async with context['dbpool'].acquire() as conn:
            j = (kernels.join(groups, groups.c.id == kernels.c.group_id)
                        .join(users, users.c.uuid == kernels.c.user_uuid))
            query = (
                sa.select([kernels, groups.c.name, users.c.email])
                .select_from(j)
                .where(
                    (kernels.c.access_key.in_(access_keys)) &
                    (kernels.c.cluster_role == DEFAULT_ROLE)
                )
                .order_by(
                    sa.desc(sa.func.greatest(
                        kernels.c.created_at,
                        kernels.c.terminated_at,
                        kernels.c.status_changed,
                    ))
                )
                .limit(100))
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if group_id is not None:
                query = query.where(kernels.c.group_id == group_id)
            if status is not None:
                query = query.where(kernels.c.status == status)
            return await batch_result(
                context, conn, query, cls,
                access_keys, lambda row: row['access_key'],
            )

    @classmethod
    async def batch_load_detail(cls, context, sess_ids, *,
                                domain_name=None, access_key=None,
                                status=None):
        async with context['dbpool'].acquire() as conn:
            status_list = []
            if isinstance(status, str):
                status_list = [KernelStatus[s] for s in status.split(',')]
            elif isinstance(status, KernelStatus):
                status_list = [status]
            elif status is None:
                status_list = [KernelStatus['RUNNING']]
            j = (kernels.join(groups, groups.c.id == kernels.c.group_id)
                        .join(users, users.c.uuid == kernels.c.user_uuid))
            query = (sa.select([kernels, groups.c.name, users.c.email])
                       .select_from(j)
                       .where((kernels.c.cluster_role == DEFAULT_ROLE) &
                              (kernels.c.session_id.in_(sess_ids))))
            if domain_name is not None:
                query = query.where(kernels.c.domain_name == domain_name)
            if access_key is not None:
                query = query.where(kernels.c.access_key == access_key)
            if status_list:
                query = query.where(kernels.c.status.in_(status_list))
            return await batch_multiresult(
                context, conn, query, cls,
                sess_ids, lambda row: row['sess_id'],
            )


class LegacyComputeSessionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(LegacyComputeSession, required=True)


async def recalc_concurrency_used(db_conn: SAConnection, access_key: AccessKey) -> None:
    query = (
        sa.update(keypairs)
        .values(
            concurrency_used=(
                sa.select([sa.func.count(kernels.c.id)])
                .select_from(kernels)
                .where(
                    (kernels.c.access_key == access_key) &
                    (kernels.c.status.in_(USER_RESOURCE_OCCUPYING_KERNEL_STATUSES))
                )
                .as_scalar()
            ),
        )
        .where(keypairs.c.access_key == access_key)
    )
    await db_conn.execute(query)
